@@ -25,7 +25,13 @@ pub fn derive(node: &DeriveInput) -> Result<TokenStream> {
 
 fn derive_struct(node: &DeriveInput, strut: &DataStruct) -> Result<TokenStream> {
     match &strut.fields {
-        Fields::Named(fields) => derive_c_struct(node, strut, fields),
+        Fields::Named(fields) => {
+            if common::has_untagged(&node.attrs) {
+                derive_untagged_c_struct(node, strut, fields)
+            } else {
+                derive_c_struct(node, strut, fields)
+            }
+        }
         Fields::Unnamed(fields) => {
             let len = fields.unnamed.len();
             match len {
@@ -91,9 +97,9 @@ fn derive_c_struct(
             filters.push(quote! {
                 #tag => {
                     if #ident.is_some() {
-                        return Err(::msgpack_schema::DeserializeError::DuplicatedField);
+                        return Err(::msgpack_schema::InvalidInputError.into());
                     }
-                    #ident = Some(::msgpack_schema::Deserialize::deserialize(__deserializer)?);
+                    #ident = Some(__deserializer.deserialize()?);
                 }
             });
         }
@@ -106,7 +112,7 @@ fn derive_c_struct(
                 }
             } else {
                 quote! {
-                    #ident: #ident.ok_or(::msgpack_schema::DeserializeError::MissingField)?,
+                    #ident: #ident.ok_or(::msgpack_schema::ValidationError)?,
                 }
             };
             ctors.push(push);
@@ -114,17 +120,17 @@ fn derive_c_struct(
 
         quote! {
             let __len = __deserializer
-                .deserialize()?
+                .deserialize_token()?
                 .to_map()
-                .ok_or(::msgpack_schema::DeserializeError::InvalidType)?;
+                .ok_or(::msgpack_schema::ValidationError)?;
 
             #( #init )*
             for _ in 0..__len {
-                let __tag: u32 = ::msgpack_schema::Deserialize::deserialize(__deserializer)?;
+                let __tag: u32 = __deserializer.deserialize()?;
                 match __tag {
                     #( #filters )*
                     _ => {
-                        let ::msgpack_schema::value::Any = ::msgpack_schema::Deserialize::deserialize(__deserializer)?;
+                        let ::msgpack_schema::value::Any = __deserializer.deserialize()?;
                     }
                 }
             }
@@ -137,10 +143,7 @@ fn derive_c_struct(
     let gen = quote! {
         #[allow(unused_qualifications)]
         impl #impl_generics #deserialize_trait for #ty #ty_generics #where_clause {
-            fn deserialize<D>(__deserializer: &mut D) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError<D::Error>>
-            where
-                D: ::msgpack_schema::Deserializer,
-            {
+            fn deserialize(__deserializer: &mut ::msgpack_schema::Deserializer) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError> {
                 #fn_body
             }
         }
@@ -159,16 +162,13 @@ fn derive_newtype_struct(
     let deserialize_trait = spanned_deserialize_trait(node);
 
     let fn_body = quote! {
-        ::msgpack_schema::Deserialize::deserialize(__deserializer).map(Self)
+        __deserializer.deserialize().map(Self)
     };
 
     let gen = quote! {
         #[allow(unused_qualifications)]
         impl #impl_generics #deserialize_trait for #ty #ty_generics #where_clause {
-            fn deserialize<D>(__deserializer: &mut D) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError<D::Error>>
-            where
-                D: ::msgpack_schema::Deserializer,
-            {
+            fn deserialize(__deserializer: &mut ::msgpack_schema::Deserializer) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError> {
                 #fn_body
             }
         }
@@ -201,7 +201,7 @@ fn derive_enum(node: &DeriveInput, enu: &DataEnum) -> Result<TokenStream> {
                             clauses.push(quote! {
                                 #tag => {
                                     if __is_array {
-                                        return Err(::msgpack_schema::DeserializeError::InvalidType);
+                                        return Err(::msgpack_schema::ValidationError.into());
                                     }
                                     Ok(Self::#ident())
                                 }
@@ -211,9 +211,9 @@ fn derive_enum(node: &DeriveInput, enu: &DataEnum) -> Result<TokenStream> {
                             clauses.push(quote! {
                                 #tag => {
                                     if !__is_array {
-                                        return Err(::msgpack_schema::DeserializeError::InvalidType);
+                                        return Err(::msgpack_schema::ValidationError.into());
                                     }
-                                    Ok(Self::#ident(Deserialize::deserialize(__deserializer)?))
+                                    Ok(Self::#ident(__deserializer.deserialize()?))
                                 }
                             });
                         }
@@ -236,23 +236,23 @@ fn derive_enum(node: &DeriveInput, enu: &DataEnum) -> Result<TokenStream> {
         }
 
         quote! {
-            let (__tag, __is_array): (u32, bool) = match __deserializer.deserialize()? {
+            let (__tag, __is_array): (u32, bool) = match __deserializer.deserialize_token()? {
                 ::msgpack_schema::Token::Int(v) => {
-                    (<u32 as ::std::convert::TryFrom<_>>::try_from(v).map_err(|_| ::msgpack_schema::DeserializeError::IntegerOutOfRange)?, false)
+                    (<u32 as ::std::convert::TryFrom<_>>::try_from(v).map_err(|_| ::msgpack_schema::ValidationError)?, false)
                 }
                 ::msgpack_schema::Token::Array(len) => {
                     if len != 2 {
-                        return Err(::msgpack_schema::DeserializeError::InvalidLength);
+                        return Err(::msgpack_schema::ValidationError.into());
                     }
-                    (u32::deserialize(__deserializer)?, true)
+                    (__deserializer.deserialize::<u32>()?, true)
                 }
                 _ => {
-                    return Err(::msgpack_schema::DeserializeError::InvalidType);
+                    return Err(::msgpack_schema::ValidationError.into());
                 }
             };
             match __tag {
                 #( #clauses )*
-                _ => Err(::msgpack_schema::DeserializeError::UnknownVariant),
+                _ => Err(::msgpack_schema::ValidationError.into()),
             }
         }
     };
@@ -260,10 +260,7 @@ fn derive_enum(node: &DeriveInput, enu: &DataEnum) -> Result<TokenStream> {
     let gen = quote! {
         #[allow(unused_qualifications)]
         impl #impl_generics #deserialize_trait for #ty #ty_generics #where_clause {
-            fn deserialize<D>(__deserializer: &mut D) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError<D::Error>>
-            where
-                D: ::msgpack_schema::Deserializer,
-            {
+            fn deserialize(__deserializer: &mut ::msgpack_schema::Deserializer) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError> {
                 #fn_body
             }
         }
@@ -318,26 +315,85 @@ fn derive_untagged_enum(node: &DeriveInput, enu: &DataEnum) -> Result<TokenStrea
             let ident = variant.ident.clone();
             let ty = field.ty.clone();
             clauses.push(quote! {
-                if let Ok(x) = ::msgpack_schema::value::deserialize::<#ty>(__value.clone()) {
+                if let Some(x) = __deserializer.try_deserialize::<#ty>()? {
                     return Ok(Self::#ident(x));
                 }
             })
         }
 
         quote! {
-            let __value = ::msgpack_schema::value::Value::deserialize(__deserializer)?;
             #( #clauses )*
-            Err(::msgpack_schema::DeserializeError::UnknownVariant)
+            Err(::msgpack_schema::ValidationError.into())
         }
     };
 
     let gen = quote! {
         #[allow(unused_qualifications)]
         impl #impl_generics #deserialize_trait for #ty #ty_generics #where_clause {
-            fn deserialize<D>(__deserializer: &mut D) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError<D::Error>>
-            where
-                D: ::msgpack_schema::Deserializer,
-            {
+            fn deserialize(__deserializer: &mut ::msgpack_schema::Deserializer) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError> {
+                #fn_body
+            }
+        }
+    };
+
+    Ok(gen)
+}
+
+fn derive_untagged_c_struct(
+    node: &DeriveInput,
+    _strut: &DataStruct,
+    named_fields: &FieldsNamed,
+) -> Result<TokenStream> {
+    let ty = &node.ident;
+    let (impl_generics, ty_generics, where_clause) = node.generics.split_for_impl();
+    let deserialize_trait = spanned_deserialize_trait(node);
+
+    let fn_body = {
+        let mut members = vec![];
+        for field in &named_fields.named {
+            let ident = field.ident.clone().unwrap();
+            let ty = field.ty.clone();
+            members.push((ident, ty))
+        }
+
+        let len = members.len() as u32;
+
+        let mut init = vec![];
+        for (ident, ty) in &members {
+            let push = quote! {
+                let mut #ident: #ty = __deserializer.deserialize()?;
+            };
+            init.push(push);
+        }
+
+        let mut ctors = vec![];
+        for (ident, _ty) in &members {
+            let push = quote! {
+                #ident,
+            };
+            ctors.push(push);
+        }
+
+        quote! {
+            let __len = __deserializer
+                .deserialize_token()?
+                .to_array()
+                .ok_or(::msgpack_schema::ValidationError)?;
+
+            if __len != #len {
+                return Err(::msgpack_schema::ValidationError.into());
+            }
+            #( #init )*
+            Ok(Self {
+                #( #ctors )*
+            })
+        }
+    };
+
+    let gen = quote! {
+        #[allow(unused_qualifications)]
+        impl #impl_generics #deserialize_trait for #ty #ty_generics #where_clause {
+            fn deserialize(__deserializer: &mut ::msgpack_schema::Deserializer) -> ::std::result::Result<Self, ::msgpack_schema::DeserializeError> {
                 #fn_body
             }
         }
